package tlschannel.async

import java.nio.ByteBuffer
import java.util.concurrent.TimeUnit

import org.scalatest.Assertions
import org.scalatest.funsuite.AnyFunSuite

import tlschannel.helpers.AsyncSocketPair
import tlschannel.helpers.SocketPairFactory
import tlschannel.helpers.SslContextFactory

class AsyncShutdownTest extends AnyFunSuite with Assertions with AsyncTestBase {

  val sslContextFactory = new SslContextFactory
  val factory = new SocketPairFactory(sslContextFactory.defaultContext)

  val bufferSize = 10

  test("immediate shutdown") {
    val channelGroup = new AsynchronousTlsChannelGroup()
    val socketPairCount = 100
    val socketPairs = factory.asyncN(null, channelGroup, socketPairCount, runTasks = true)
    for (AsyncSocketPair(client, server) <- socketPairs) {
      val writeBuffer = ByteBuffer.allocate(bufferSize)
      client.external.write(writeBuffer)
      val readBuffer = ByteBuffer.allocate(bufferSize)
      server.external.read(readBuffer)
    }

    assert(!channelGroup.isTerminated)

    channelGroup.shutdownNow()

    // terminated even after a relatively short timeout
    val terminated = channelGroup.awaitTermination(100, TimeUnit.MILLISECONDS)
    assert(terminated)
    assert(channelGroup.isTerminated)
    assertChannelGroupConsistency(channelGroup)

    printChannelGroupStatus(channelGroup)
  }

  test("non-immediate shutdown") {
    val channelGroup = new AsynchronousTlsChannelGroup()
    val socketPairCount = 100
    val socketPairs = factory.asyncN(null, channelGroup, socketPairCount, runTasks = true)
    for (AsyncSocketPair(client, server) <- socketPairs) {
      val writeBuffer = ByteBuffer.allocate(bufferSize)
      client.external.write(writeBuffer)
      val readBuffer = ByteBuffer.allocate(bufferSize)
      server.external.read(readBuffer)
    }

    assert(!channelGroup.isTerminated)

    channelGroup.shutdown()

    {
      // not terminated even after a relatively long timeout
      val terminated = channelGroup.awaitTermination(2000, TimeUnit.MILLISECONDS)
      assert(!terminated)
      assert(!channelGroup.isTerminated)
    }

    for (AsyncSocketPair(client, server) <- socketPairs) {
      client.external.close()
      server.external.close()
    }

    {
      // terminated even after a relatively short timeout
      val terminated = channelGroup.awaitTermination(100, TimeUnit.MILLISECONDS)
      assert(terminated)
      assert(channelGroup.isTerminated)
    }

    assertChannelGroupConsistency(channelGroup)

    assert(channelGroup.getCancelledReadCount == 0)
    assert(channelGroup.getCancelledWriteCount == 0)
    assert(channelGroup.getFailedReadCount == 0)
    assert(channelGroup.getFailedWriteCount == 0)

    printChannelGroupStatus(channelGroup)
  }

}
